#ifndef auto_rfutils_h
#define auto_rfutils_h 1

#include "AutoRandomFieldsUtilsLocal.h"

#define MAXUNITS 4
#define MAXCHAR 18 // max number of characters for (covariance) names  
#define RFOPTIONS "RFoptions"
#define isGLOBAL NA_INTEGER

#endif
